# -*- encoding: utf-8 -*-

"""
A List of Useful Functions to Provide Functionality over a Dataframe

The :attr:`numpy` and :attr:`pandas` has various in-built functions
to quickly summarize dataframe. The functions tabs enables end-users
with the capability of creating custom functions (like half-life) and
apply the same over a pandas dataframe object.
"""

from pandaswizard.functions.collate import * # noqa: F401, F403 # pyright: ignore[reportMissingImports]



# -*- encoding: utf-8 -*-

"""
A Utility Wrapper for the Rolling Functions for the :attr:`pandas` Module

The :attr:`pandas.core.window.rolling()` is a useful window function
that can be used to capture historic data movement patterns (like
moving averages). However, there are some limitations like application
of user defined function to the rolling module. For example, we may
want to average out based on "weighted moving average" function.
"""

import numpy as np
import pandas as pd

from typing import Union, Callable

def _rolling(series : pd.Series, window : Union[int, Callable], **kwargs):
    """
    Wrapper Function that Syncs the Method with :attr:`.rolling()` Method

    The wrapper function accepts all the keyword arguments and returns
    only the arguments with their default value (or user-defined) to
    the :attr:`rolling()` method.

    Limitations: the wrapper function will work only with a data
    of type :attr:`pd.Series` and thus, does not supporr the :attr:`on`
    parameter which selects column from the data frame. The defined
    wrapper will typically work like:

    .. code-block:: python
        data = pd.DataFrame(data = {...})
        data["calculations"] = data["column"].apply(pdw.rolling(...))
    """

    min_periods = kwargs.get("min_periods", None)
    center = kwargs.get("center", False)
    win_type = kwargs.get("win_type", None)
    closed = kwargs.get("closed", None)
    step = kwargs.get("step", None)

    return series.rolling(
        window,
        min_periods = min_periods,
        center = center,
        win_type = win_type,
        closed = closed,
        step = step
    )


def rolling(
        series : Union[pd.Series, np.ndarray],
        window : Union[int, Callable],
        method : Callable = np.mean,
        **kwargs
    ) -> np.ndarray:
    """
    A Wrapper for the :attr:`pd.DataFrame(...).rolling()` Module

    The :attr:`pandas` module useful window function :attr:`rolling`
    which are most of the time sufficient for aggregating over a
    data series but limited to in-built functions, or a custom function
    can be applied by using `.apply(...)` over the rolling values.
    To address this issue, the function wraps the method and rolling
    calculations in one place and returns an array to be directly
    added to the dataframe. Example is as below:

    .. code-block:: python
        data = pd.DataFrame(np.random.random((100, 1)), columns = ["A"])

        data["MA-50"] = data["A"].rolling(50).mean() # using rolling method

    Instead of moving average, we can also give more weightage to the
    recent time-period (weighted average) like:

    .. code-block:: python
        udf = lambda x : (x * np.array([0.125, 0.250, 0.500])).sum()
        data["wma"] = data["A"].rolling(3).apply(udf)

    To reduce iterative process of defining the factors (of :attr:`udf`)
    and creating one step calculations this function can be helpful.
    The revised syntax is like:

    .. code-block:: python
        data["wma"] = pdw.rolling(data["A"], 50, method = np.mean, ...)

    The following paramteres and keyword arguments are accepted by the
    function to provide dynamic approach as below:

    :type  series: object
    :param series: The dataframe series object which can be typically
        accessed directly like :attr:`dataframe["column"]` or an
        n-dimensional array like :attr:`dataframe["column"].values`.

    :type  window: int | callable
    :param window: The length of the window which is either an integer
        or a callable value which is supported by the
        `pd.DataFrame.rolling` function.

    :type  method: callable
    :param method: The aggregation method of the rolling series,
        which can be a simple function like :attr:`np.mean` or
        can be any of the custom fancy functions.
    """

    rolling_ = _rolling(series, window, **kwargs) # rolling object
    return [
        method(array) if array.shape[0] == window
        else np.nan for array in rolling_
    ]



# -*- encoding: utf-8 -*-

"""
A Set of Function(s) to Collate/Aggregate based on Logic

A simple example of aggregation is the statistical measures like
:attr:`mean`, :attr:`median`, etc. Additional in-frequently but
popular aggregation functions are defined here for end-users.
"""

import numpy as np
from typing import Union, Callable

def weightedMA(initial : float, rate : Union[float, Callable], length : int, decay : bool = True) -> np.ndarray:
    """
    Collate a Series based on Weighted Moving Average (WMA) Method

    WMA is a variant of SMA/EMA, and is popularly used in financial
    analysis, which gives more weightage to the recent data and
    produces a smoother line (sometimes) giving a more accurate picture
    of the underlying data trend.

    :type  initial: float
    :param initial: The initial weighteage of the value, typically
        a value of :attr:`0.5` is a good starting point.

    :type  rate: float, callable
    :param rate: The rate at which subsequent values are increasing
        or decreasing. Typically, a value of :attr:`2` (i.e., at each
        subsequent level the impact is halved - "half life decay") is
        a good starting point. The rate can either be a numeric value,
        i.e., each subsequent values is calculated as
        :attr:`n_1 = n_0 / rate` or can be a callable, i.e., each
        value is calculated like :attr:`n_1 = rate(n_0)` thus allow
        more control and dynamic approach.

    :type  length: int
    :param length: Length of the window, this enables a quick
        summarization of the final outcome using `x * weightedMA()`,
        where :attr:`x` is also a n-dimensional :attr:`numpy` array.

    :type  decay: bool
    :param decay: When true (default) the returned array will be
        reveresed, i.e., it will give more priority to the recent
        data points (where the :attr:`x` is sorted in ascending order),
        else typically returns a "growth" array where more weightage
        is given to the data which is older.
    """

    factors = [initial] # append the initial values, and then calculate
    for _ in range(length - 1):
        factors.append(
            rate(factors[-1]) if hasattr(rate, "__call__")
            else factors[-1] / rate
        )

    factors = np.array(factors)
    return factors[::-1] if decay else factors



# -*- encoding: utf-8 -*-

"""
A List of Useful Decorators/Wrappers for `pandas` Modules

The decorators can be used for profiling/understanding the a function
which are developed to handle a `DataFrame` object.
"""

import time
import functools
import collections
import pandas as pd

def recordCounter(func : callable) -> callable:
    """
    Verbose the Shape of the DataFrame Pre- & Post- Function Execution

    On execution of a function (typically, if the function uses the
    :attr:`.drop()` or :attr:`.merge()` or etc.) the decorator prints
    the number of records before and after the execution.

    .. code-block:: python

        import pandas as pd
        import pandaswizard as pdw # attempt to create an ubiquitous naming

        frame = pd.DataFrame(data = {
            "LABEL" : ["A", "A", "B"],
            "VALUES" : [1, 2, 3]
        })

        @pdw.wrappers.recordCounter
        def dropvals(frame):
            return frame[frame["LABEL"] != "B"]

        filtered = dropvals(frame = frame.copy())
        >> Executed with @recordCounter[`dropvals`]
        >>   >> Original Record Count = 3
        >>   >> Final Record Count    = 2
        >>   >> Dropped/Added Records = -1 (= 33.333%)

    LIMITATION: The decorator works IFF the dataframe is either the
    first arguments or is passed as an keyword argument with the
    argument name as either `data` or `frame` which is the general
    convention that is followed throughout the module. If the code has
    arguments, then it has preference over the keyword arguments and
    the `dataframe` must be the first argument.

    USE Cases: Often on function execution, joining/dropping records
    users need to execute `df.shape` before and after to check the
    record count. This is simplified and gives the result, and thus
    reducing code during such operations.
    """

    @functools.wraps(func)
    def _wrapper(*args, **kwargs) -> pd.DataFrame:
        print(f"Executed with @recordCounter[`{func.__name__}`]")

        errors = False # lets initialize the decorator w/o error
        try:
            __frame = args[0] if args else \
                kwargs.get("data", None) or kwargs["frame"]
        except KeyError:
            errors = True
            print("  >> Failed to Execute. Check Decorator Limitation.")

        if not errors and not isinstance(__frame, pd.DataFrame):
            errors = True
            print("  >> Failed to Execute. DF Object is Not Found.")

        start_record_count_ = __frame.shape[0] if not errors else 0
        retval = func(*args, **kwargs) # ? execute the func as is

        # ! retval:: can either be a dataframe, or iterable where
        # the first returned value is always a dataframe, then we can
        __retframe = retval[0] if isinstance(retval, (list, tuple)) \
            else retval if isinstance(retval, pd.DataFrame) else None

        try:
            final_record_count_ = __retframe.shape[0]
        except Exception as err:
            errors, final_record_count_ = True, 0
            print(f"  >> Function does not Return a DF. ERROR: {err}")

        if not errors:
            print(f"  >> Original Record Count = {start_record_count_:,}")
            print(f"  >> Final Record Count    = {final_record_count_:,}")

            n_change = final_record_count_ - start_record_count_
            p_change = (abs(n_change) / start_record_count_) * 100 if not errors else 0
            print(f"  >> Dropped/Added Records = {n_change:,} (= {p_change:.3f}%)")

        return retval

    return _wrapper


def timeit(func : callable) -> callable:
    """
    A Mimic of the iPython Magic :attr:`%timeit` for Dataframes

    The built-in iPython magic function :attr:`%timeit` displays the
    executed time of a function, or like the one from command line:

    .. code-block:: shell

        python -m timeit "function()"

    The function is built specifically to handle functions which
    returns a :attr:`pd.DataFrame` object, thus it prints more
    information like the number of records fetched, shape and other
    information w/o explictly needing to call additional pandas
    function by the end-user.
    """

    @functools.wraps(func)
    def _wrapper(*args, **kwargs) -> pd.DataFrame:
        print(f"Executed with @timeit[`{func.__name__}`]")

        start = time.process_time() # capture start time
        frame = func(*args, **kwargs) # execute function

        # verbose information(s) to the output
        process_time = time.process_time() - start
        dtypes_count = collections.Counter(frame.dtypes.values)
        print(f"  >> Function Executed in {process_time:,.3f} secs.")
        print(f"  >> Fetched {frame.shape[0]:,} Record(s).")
        print(f"  >> No. of Feature(s)/Column(s) = {frame.shape[1]:,}.")
        print(f"  >> Observed Data Type(s) : {dtypes_count}")

        return frame
    return _wrapper



#!/usr/bin/env python
#
# Copywright (C) 2024 Debmalya Pramanik <neuralNOD@gmail.com>
# LICENSE: MIT License

from setuptools import setup
from setuptools import find_packages

import pandaswizard as pdw

setup(
    name = "pandas-wizard",
    version = pdw.__version__,
    author = "shark-utilities developers",
    author_email = "neuralNOD@outlook.com",
    description = "Utility Functions, Wrappers for pandas Module",
    long_description = open("README.md", "r").read(),
    long_description_content_type = "text/markdown",
    url = "https://github.com/sharkutilities/pandas-wizard",
    packages = find_packages(),
    classifiers = [
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: MIT License"
    ],
    project_urls = {
        "Issue Tracker" : "https://github.com/sharkutilities/pandas-wizard/issues",
        "Code Documentations" : "https://pandas-wizard.readthedocs.io/en/latest/index.html",
        "Org. Homepage" : "https://github.com/sharkutilities"
    },
    keywords = [
        # keywords for finding the package::
        "pandas", "utility", "utilities", "util", "utils",
        # keywords for finding the package relevant to usecases::
        "wrappers", "data science", "data analysis", "data scientist", "data analyst",
        # keywords as per available functionalities::
        "statistics", "percentile", "quantile", "probability"
    ],
    python_requires = ">=3.8"
)



# -*- encoding: utf-8 -*-

"""
Utility Function(s) Related to a Grouped/Aggregated Data Frame

The basic syntax for the groupby aggregation is `pd.groupby().agg({})`
and the utility functions provided here can be applied under the
aggregation section.
"""

import warnings
import numpy as np
import pandas as pd

def __set_method__(kwargs : dict) -> str:
    """
    Set "Method"/"Interpolation" Attribute for Aggregated Function(s)

    Check individual function for attribute defination and usages.
    The function returns values as used by :func:`percentile` and
    :func:`quantile` methods defined below.
    """

    assert not all(["method" in kwargs, "interpolation" in kwargs]), \
        "Either `method` or `interpolation` is required. Received both."

    method = kwargs.get("method", kwargs.get("interpolation", "linear"))
    return method


def __set_basemod__(basemod : str) -> str:
    assert basemod in ["pd", "pandas", "np", "numpy"], \
        f"basemod = {basemod} is not valid, and/or not implemented."

    basemod = "pd" if basemod in ["pd", "pandas"] else "np"
    return basemod


def __calculate_quantile__(
            x : pd.Series,
            n : float,
            method : str,
            func : str,
            dropna : bool,
            basemod : str
    ) -> float:
    """
    Calculates Percentile/Quantile for a Grouped Series

    The function can calculate both percentile and/or quantile, but
    name is set as `__calculate_quantile__()` as default. Uses either
    the `numpy` or the `pandas` module to calculate the result.
    """

    retval = None # ? return value, i.e., quantile/percentile
    basemod = __set_basemod__(basemod) # ? defaults to pd (pandas)

    if basemod == "pd":
        # ? use pandas to calculate the series quantile/percentile
        # this is the default feature, and mimics np.nanquantile
        retval = x.quantile(n, interpolation = method)
    else:
        # ? else use numpy to calculate series quantile/percentile
        # ! this is always true, asserted in `__set_basemod__()`
        x = x.values # ? convert to an np.ndarray
        n = n * 100 if func == "percentile" else n

        __func_dispatcher__ = {
            "percentile" : {
                True : np.nanpercentile,
                False : np.percentile
            },
            "quantile" : {
                True : np.nanquantile,
                False : np.quantile
            }
        }

        try:
            retval = __func_dispatcher__[func][dropna](x, n, method = method)
        except TypeError as err:
            __ref_issue = "https://github.com/numpy/numpy/issues/21283"
            warnings.warn(
                f"NumPy/np Version < 1.22, {__ref_issue} Error: {err}",
                FutureWarning
            )

            # ? for older version, try with attribute `interpolation`, else
            # ! this should not raise error, unless legacy numpy version, OR
            # method/interpolation value is given for a newer version, not available
            try:
                retval = __func_dispatcher__[func][dropna](x, n, interpolation = method)
            except Exception as err:
                __warn_message = f"Cannot call attribute `method/interpolation` = {err}"
                warnings.warn(f"{__warn_message}. Returning value w/o argument", SyntaxWarning)
                retval = __func_dispatcher__[func][dropna](x, n)

    return retval


def percentile(n : float, outname : str = None, **kwargs) -> float:
    """
    Compute the n-th Percentile for the Grouped Data Series

    In statistics, a `n`-th percentile, also known as centile score,
    is a score below which a given percentage `n` of scores in its
    frequency distribution falls or a score at or below which the
    given percentage falls. More information is available
    [here](https://en.wikipedia.org/wiki/Percentile). Percentiles are
    a type of [*quanitile*](https://en.wikipedia.org/wiki/Quantile)
    and can be interchangeably used.

    Internally, the function uses the :meth:`pd.Series.quantile`
    method to calculate the n-th percentile of the grouped series.

    :type  n: int or float
    :param n: Percentage value to compute. Values must be between
        `[0, 100]` both inclusive.

    :type  outname: str
    :param outname: Output name of the aggregated feature when the
        method is used in conjuncture with other functions. This
        does not have any significance when used as in the below
        example. The outname defaults to :attr:`f"P{n:.2f}"`.

    Keyword Arguments
    -----------------
        * **method** (*str*): This parameter specifies the method to
            use for estimating the percentile. There are many
            different methods of which some are unique to NumPy.
            Accepts any value as in :meth:`np.percentile`
            parameter, defaults to "linear" method. However, for
            the `pd.Series.quantile()` the argument `method` is
            termed as `interpolation` and the values can be:
            {'linear', 'lower', 'higher', 'midpoint', 'nearest'}.

        * **interpolation** (*str*): Same as :attr:`method` the
            method for quantile calculation as per pandas. Both the
            attribute :attr:`method` and :attr:`interpolation` cannot
            be passed at the same time, and raises `AssertionError`
            if done so.

        * **basemod** (*str*): Abbreviation for "base module", allows
            the user to choose from `pandas` or `numpy` to calculate
            percentile. When choosing `numpy` the default behaviour
            is `np.nanquantile()` as followed by `pd.Series.quantile`
            however, you can pass `dropna = False` which calculates
            using `np.percentile` and returns `np.nan` if input
            contain nan values. Defaults to `pandas`. Allowed terms:
            {'pd', 'pandas', 'np', 'numpy'}.

        * **dropna** (*bool*): Calculate the percentile by dropping
            the `nan` values. This method mimics the `np.nanpercentile`
            function, which is the default as in `pd.Series.quantile()`.
            More information: https://stackoverflow.com/a/70002786.

    Example and Usages
    ------------------

    Assuming an end-user have the basic understanding of `pandas` and
    `percentile`, we can use compute the percentile for a group like:

    .. code-block:: python

        import pandas as pd
        import pandaswizard as pdw

        data = pd.DataFrame({"G" : ["A", "B", "B"], "V" : [1, 2, 3]})

        # CASE-I: standalone usage, can be used on multiple features
        percentile = data.groupby("A").agg("V" : pdw.percentile(50))

        # CASE-II: usage in conjunture with any other function
        # `.agg({})` passing dictionary of values, or by passing
        # named tuples like `.agg(outname = ("column" : ))`
        # more details: https://stackoverflow.com/a/53619715/6623589

        percentile = data.groupby("A").agg({
            "V" : [sum, pdw.percentile(50, outname = "P50")]
        })

    Both the methods calculates the percentile for the grouped value.
    In **CASE-I** the argument "outname" does not have any implications
    as `pandas` by default returns using the result with the original
    name, however in case of **CASE-II** we can set the feature name
    using the argument `outname`.
    """

    method = __set_method__(kwargs)

    # ? the validation and check are done in __calculate_quantile__()
    dropna = kwargs.get("dropna", True)
    basemod = kwargs.get("basemod", "pandas")

    def percentile_(x : list) -> float:
        return __calculate_quantile__(x, n = n / 100, method = method, func = "percentile", dropna = dropna, basemod = basemod)

    percentile_.__name__ = outname or f"P{n:.2f}"
    return percentile_


def quantile(n : float, outname : str = None, **kwargs) -> float:
    """
    Compute the n-th Quantile for the Grouped Data Series

    In statisticsand probability, quantiles are cut points dividing
    the range of probability distribution into continuous intervals
    with equal probabilities, or dividing the observation in a
    sample in the same way. More information is available
    [here](https://en.wikipedia.org/wiki/Quantile).

    Internally, the function uses the :meth:`pd.Series.quantile`
    to calculate the n-th quantile of the grouped series.

    :type  n: int or float
    :param n: Probability value for the quantiles to compute. The
        values must be between `[0, 1]` both inclusive.

    :type  outname: str
    :param outname: Output name of the aggregated feature when the
        method is used in conjuncture with other functions. This
        does not have any significance when used as in the below
        example. The outname defaults to :attr:`f"Q{n:.2f}"`.

    Keyword Arguments
    -----------------
        * **method** (*str*): This parameter specifies the method to
            use for estimating the quantile. There are many
            different methods of which some are unique to NumPy.
            Accepts any value as in :meth:`np.quantile`
            parameter, defaults to "linear" method. However, for
            the :meth:`pd.Series.quantile` the argument `method` is
            termed as `interpolation` and the values can be:
            {'linear', 'lower', 'higher', 'midpoint', 'nearest'}.

        * **interpolation** (*str*): Same as :attr:`method` the
            method for quantile calculation as per pandas. Both the
            attribute :attr:`method` and :attr:`interpolation` cannot
            be passed at the same time, and raises `AssertionError`
            if done so.

        * **basemod** (*str*): Abbreviation for "base module", allows
            the user to choose from `pandas` or `numpy` to calculate
            percentile. When choosing `numpy` the default behaviour
            is :meth:`np.nanquantile` as followed by `pd.Series.quantile`
            however, you can pass `dropna = False` which calculates
            using `np.percentile` and returns `np.nan` if input
            contain nan values. Defaults to `pandas`. Allowed terms:
            {'pd', 'pandas', 'np', 'numpy'}.

        * **dropna** (*bool*): Calculate the percentile by dropping
            the `nan` values. This method mimics the `np.nanpercentile`
            function, which is the default as in `pd.Series.quantile()`.
            More information: https://stackoverflow.com/a/70002786.

    Example and Usages
    ------------------

    Assuming an end-user have the basic understanding of `pandas` and
    `quantile`, we can use compute the quantile for a group like:

    .. code-block:: python

        import pandas as pd
        import pandaswizard as pdw

        data = pd.DataFrame({"G" : ["A", "B", "B"], "V" : [1, 2, 3]})

        # CASE-I: standalone usage, can be used on multiple features
        quantile = data.groupby("A").agg("V" : pdw.quantile(50))

        # CASE-II: usage in conjunture with any other function
        # `.agg({})` passing dictionary of values, or by passing
        # named tuples like `.agg(outname = ("column" : ))`
        # more details: https://stackoverflow.com/a/53619715/6623589

        quantile = data.groupby("A").agg({
            "V" : [sum, pdw.quantile(50, outname = "Q0.5")]
        })

    Both the methods calculates the quantile for the grouped value.
    In **CASE-I** the argument "outname" does not have any implications
    as `pandas` by default returns using the result with the original
    name, however in case of **CASE-II** we can set the feature name
    using the argument `outname`.
    """

    method = __set_method__(kwargs)

    # ? the validation and check are done in __calculate_quantile__()
    dropna = kwargs.get("dropna", True)
    basemod = kwargs.get("basemod", "pandas")

    def quantile_(x : list) -> float:
        return __calculate_quantile__(x, n = n, method = method, func = "quantile", dropna = dropna, basemod = basemod)

    quantile_.__name__ = outname or f"Q{n * 100:.2f}"
    return quantile_



# -*- encoding: utf-8 -*-

"""
Utility Functions and/or Wrappers for pandas Library

The :mod:`pandaswizard` module includes functions, wrappers and other
utility functions for `pandas` module. The package is kept simple and
minimalistic such that external dependencies are reduced. The working
of the module is divided into the following sections:

    * :mod:`pandaswizard.aggregate`: a set of aggregate functions
      that can be used along with ``pd.groupby().agg({...})`` method
      without comprimising functionality.
    * :mod:`pandaswizard.wrappers`: a set of decorators/wrappers
      that can be used along side a function.
"""

# ? package follows https://peps.python.org/pep-0440/
# ? https://python-semver.readthedocs.io/en/latest/advanced/convert-pypi-to-semver.html
__version__ = "1.1.0"

# init-time options registrations
from pandaswizard.aggregate import (
    quantile,
    percentile
)

from pandaswizard import window
from pandaswizard import wrappers
from pandaswizard import functions
